/**
 * netica_pronouncer.c
 * ===================
 * Implementation of basic pronouncer functionality based on Netica API.
 *
 * By Hkan L. Younes (lorens@acm.org)
 */

#include <stdlib.h>

#include <Netica.h>

#include "netica_pronouncer.h"


/**
 * Maximum number of models kept in memory.
 */
#define MAX_MODELS 10

/**
 * Maximum number of states a node can be in.
 */
#define MAX_STATES  32

/**
 * Maximum number of parents that a node can have.
 */
#define MAX_PARENTS 32


/**
 * The global "Netica environment".
 */
static environ_ns *env;

/**
 * The models.
 */
static net_bn *models[MAX_MODELS];


/**
 * Initialize pronouncer.
 *
 * Returns -1 on failure, and 0 on success.
 *
 * N.B. must be called before the pronouncer is used.
 */
int initializePronouncer(void)
{
    char msg[MESG_LEN_ns];
    int res, i;

    for (i = 0; i < MAX_MODELS; i++)
	models[i] = NULL;

    env = NewNeticaEnviron_bn(NULL);
    res = InitNetica_bn(&env, msg);

    return (res < 0) ? -1 : 0;
}


/**
 * Finalizes pronouncer.
 *
 * N.B. should be called after the last call to the pronouncer.
 */
void finalizePronouncer(void)
{
    char msg[MESG_LEN_ns];
    int i;

    for (i = 0; i < MAX_MODELS; i++)
	freeModel(i);

    CloseNetica_bn(env, msg);
}


/**
 * Loads a model.
 *
 * Parameters:
 *  path  - path to file containing the model.
 *
 * Returns -1 on failure, and an identifier for the model on success.
 */
int loadModel(const char *path)
{
    stream_ns *file;
    int i;

    for (i = 0; i < MAX_MODELS && models[i] != NULL; i++)
	;
    if (i == MAX_MODELS)
	return -1;   /* maximum number of models loaded */

    file = FileNamed_ns(path, env);
    models[i] = ReadNet_bn(file, FALSE);

    return (models[i] == NULL) ? -1 : i;
}


/**
 * Frees a model.
 *
 * Parameters:
 *  modelId - a model identifier as returned by loadModel.
 */
void freeModel(int modelId)
{
    if (models[modelId] != NULL)
	FreeNet_bn(models[modelId]);
}


int saveModel(int modelId, char* path)
{
    stream_ns *file;

    file = FileNamed_ns(path, env);
    WriteNet_bn(models[modelId], file);

    return 0;
}


/**
 * Sets probabilities in the given model.
 *
 * Parameters:
 *  modelId      - a model identifier as returned by loadModel.
 *  nodeName     - name of node.
 *  parentStates - states of parents indicating row in probability table.
 *  nprobs       - number of probabilities in probs.
 *  probs        - array of probabilities to assign to row.
 */
void setProbabilities(int modelId, const char *nodeName,
		      const int *parentStates, int nprobs,
		      const double *probs)
{
    prob_bn fprobs[MAX_STATES];
    int i;
    node_bn *node;

    for (i = 0; i < nprobs; i++)
	fprobs[i] = probs[i];
    node = NodeNamed_bn(nodeName, models[modelId]);
    SetNodeProbs_bn(node, parentStates, fprobs);
}


/**
 * Sets a utility in the given model.
 *
 * Parameters:
 *  modelId      - a model identifier as returned by loadModel.
 *  nodeName     - name of node.
 *  parentStates - states of parents indicating position in utility table.
 *  util         - the utility.
 */
void setUtility(int modelId, const char *nodeName,
		const int *parentStates, double util)
{
    node_bn *node;

    node = NodeNamed_bn(nodeName, models[modelId]);
    SetNodeFuncValue_bn(node, parentStates, util);
}


/**
 * Sets all probabilities in a node of the given model.
 *
 * Parameters:
 *  modelId       - a model identifier as returned by loadModel.
 *  nodeName      - name of node.
 *  nparents      - number of parents
 *  parentNstates - number of states that each of the parents can be in.
 *  nstates       - number of states that this node can be in.
 *  probs         - probabilities.
 */
void setAllProbabilities(int modelId, const char *nodeName,
			 int nparents, const int *parentNstates,
			 int nstates, const double *probs)
{
    int parentStates[MAX_PARENTS];
    int i;

    for (i = 0; i < nparents - 1; i++)
	parentStates[i] = 0;
    parentStates[i] = -1;
    while (i >= 0)
    {
	i = nparents - 1;
	while (++parentStates[i] == parentNstates[i] && i >= 0)
	{
	    parentStates[i] = 0;
	    i--;
	    if (i < 0)
		return;
	}
	if (i >= 0)
	{
	    setProbabilities(modelId, nodeName, parentStates, nstates, probs);
	    probs += nstates;
	}
    }
}


/**
 * Sets all utilities in a node of the given model.
 *
 * Parameters:
 *  modelId       - a model identifier as returned by loadModel.
 *  nodeName      - name of node.
 *  nparents      - number of parents
 *  parentNstates - number of states that each of the parents can be in.
 *  utils         - utilities.
 */
void setAllUtilities(int modelId, const char *nodeName, 
		     int nparents, const int *parentNstates,
		     const double *utils)
{
    int parentStates[MAX_PARENTS];
    int i;

    for (i = 0; i < nparents - 1; i++)
	parentStates[i] = 0;
    parentStates[i] = -1;
    while (i >= 0)
    {
	i = nparents - 1;
	while (++parentStates[i] == parentNstates[i] && i >= 0)
	{
	    parentStates[i] = 0;
	    i--;
	    if (i < 0)
		return;
	}
	if (i >= 0)
	{
	    setUtility(modelId, nodeName, parentStates, *utils);
	    utils++;
	}
    }
}


/**
 * Evaluates the given model.
 *
 * Parameters:
 *  modelId      - a model identifier as returned by loadModel.
 *  decisionName - name of decision node.
 *
 * Returns the index of an alternative that maximizes the expected utility.
 */
int evaluateModel(int modelId, const char *decisionName)
{
    node_bn *node;

    CompileNet_bn(models[modelId]);
    OptimizeDecisions_bn(GetNetNodes_bn(models[modelId]));
    node = NodeNamed_bn(decisionName, models[modelId]);

    return (int)GetNodeFuncValue_bn(node, NULL);
}
